/*
 * Copyright 2015 Luca Beltrame <lbeltrame@kde.org>
 *
 * This file is part of Danbooru Client.
 *
 * Danbooru Client is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Danbooru Client is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Danbooru Client. If not, see <http://www.gnu.org/licenses/>.
 */

// Own

#include "utils.h"

// Qt

#include <QJsonDocument>
#include <QXmlStreamReader>
#include <QUrlQuery>

// KDE

#include "libdanbooru_debug.h"

typedef QLatin1String QL1S;

static const QMap<QString, Danbooru::Rating> RATING_MAP = Danbooru::initRatingMap();

const QMap< QString, Danbooru::Rating > Danbooru::initRatingMap()
{

    QMap<QString, Danbooru::Rating> map;
    map.insert("s", Safe);
    map.insert("q", Questionable);
    map.insert("e", Explicit);

    return map;

}

namespace {
    QUrl adjustUrl(QUrl baseUrl, QUrl boardUrl) {
        if(baseUrl.host().isEmpty()) {
            baseUrl.setHost(boardUrl.host());
            baseUrl.setScheme(boardUrl.scheme());
        }
    return baseUrl;
    }
}

namespace Danbooru
{
    


QUrl requestUrl(QUrl &url, const QString &path,
                const QString &username, const QString &password,
                const dictMap &parameters, const QStringList &tags)
{

    QUrl danbooruUrl = QUrl(url);
    danbooruUrl = danbooruUrl.adjusted(QUrl::StripTrailingSlash);
    danbooruUrl.setPath(danbooruUrl.path() + '/' + path);

    // If we have parameters, add them

    QUrlQuery query;

    if (!parameters.isEmpty()) {

        for (auto it = parameters.cbegin(), e = parameters.cend(); it != e; ++it) {
            query.addQueryItem(it.key(), it.value());
        }

    }

    // Now, let's add tags should we have them

    if (!tags.isEmpty()) {
        query.addQueryItem("tags", tags.join(" "));
    }

    danbooruUrl.setQuery(query);

    if (!username.isEmpty() && !password.isEmpty()) {
        danbooruUrl.setUserName(username);
        danbooruUrl.setPassword(password);
    }

    return danbooruUrl;
}

QUrl requestUrl(QUrl &url, const QString &path, const QString &username,
                const QString &password, const dictMap &parameters)
{

    QUrl danbooruUrl = QUrl(url);
    danbooruUrl = danbooruUrl.adjusted(QUrl::StripTrailingSlash);
    danbooruUrl.setPath(danbooruUrl.path() + '/' + path);

    // If we have parameters, add them

    QUrlQuery query;

    if (!parameters.isEmpty()) {

        for (auto it = parameters.cbegin(), e = parameters.cend(); it != e; ++it) {
            query.addQueryItem(it.key(), it.value());
        }

    }

    danbooruUrl.setQuery(query);

    if (!username.isEmpty() && !password.isEmpty()) {
        danbooruUrl.setUserName(username);
        danbooruUrl.setPassword(password);
    }

    return danbooruUrl;
}

QUrl requestUrl(QUrl &url, const QString &path, const QString &username,
                const QString &password)
{
    QUrl danbooruUrl = QUrl(url);
    danbooruUrl = danbooruUrl.adjusted(QUrl::StripTrailingSlash);
    danbooruUrl.setPath(danbooruUrl.path() + '/' + path);

    if (!username.isEmpty() && !password.isEmpty()) {
        danbooruUrl.setUserName(username);
        danbooruUrl.setPassword(password);
    }

    return danbooruUrl;
}

QList< QVariant > parseDanbooruResult(QByteArray data, QString xlmElement, bool *result, ApiType apiType)
{

    QXmlStreamReader reader;
    reader.addData(data);
    QList<QVariant> postData;

    if (apiType != ApiType::Danbooru) {

        while (!reader.atEnd() && !reader.hasError()) {

            QXmlStreamReader::TokenType token = reader.readNext();

            if (token == QXmlStreamReader::StartDocument) {
                continue;
            }

            if (token == QXmlStreamReader::StartElement &&
                    reader.name() == xlmElement)  {

                QVariantMap values;

                QXmlStreamAttributes attributes = reader.attributes();

    //                 qCDebug(LIBDANBOORU) << attributes;

                for (auto attribute : attributes) {
                    values.insert(attribute.name().toString(),
                                attribute.value().toString());
                }

                if (values.isEmpty()) {
                    *result = false;
                    qCWarning(LIBDANBOORU) << "No results found when parsing XML";
                    return QList<QVariant>();
                }

                QVariant converted = QVariant(values);

                postData.append(converted);
            }
        }
        *result = true;
        return postData;
    } else {
        QString plural = xlmElement + QString("s");
        while(reader.readNextStartElement()) {

            if (reader.name() == xlmElement) {

                QVariantMap values;

                while(reader.readNextStartElement()) {
                    values.insert(reader.name().toString(), reader.readElementText());
                }
                QVariant converted = QVariant(values);
                postData.append(converted);
            }

        }
        if (postData.isEmpty()) {
            *result = false;
            qCWarning(LIBDANBOORU) << "No results found when parsing XML";
            return QList<QVariant>();
        }
        *result = true;
        return postData;
    }
}

QVariant parseDanbooruResult(QByteArray data, bool *result)
{

    QJsonDocument parsed = QJsonDocument::fromJson(data);

    if (parsed.isNull()) {
        return QList<QVariant>();
    }

    QVariant postData = parsed.toVariant();
    
//     qCDebug(LIBDANBOORU) << "Raw JSON response" << parsed.toJson(QJsonDocument::Indented).constData();

    *result = true;

    return postData;
}

bool isPostBlacklisted(DanbooruPost *post, QSet<QString> blacklist, Danbooru::Ratings maxRating)
{

    if (post->rating() > maxRating) {
        return true;
    }

    for (auto tag : post->tags()) {

        if (blacklist.contains(tag)) {
            return true;
        }
    }

    return false;

}

void fixPostUrl(DanbooruPost *post, QUrl boardUrl) {

    auto thumbnailUrl = post->thumbnailUrl();
    auto sampleUrl = post->sampleUrl();
    auto fileUrl = post->fileUrl();

    post->setThumbnailUrl(adjustUrl(thumbnailUrl, boardUrl));
    post->setSampleUrl(adjustUrl(sampleUrl, boardUrl));
    post->setFileUrl(adjustUrl(fileUrl, boardUrl));


}

QVariantMap extractPoolData(const QVariant &data, ApiType type) {
    
    auto mapped = data.toMap();
    int id;
    QString name;
    int postCount;
    QString description;
    QVariantList rawData;
    
    id = mapped.value("id").toInt();
    name = mapped.value("name").toString();
    postCount = mapped.value("post_count").toInt();
    description = mapped.value("description").toString();
    
    switch(type) {
        case MoeBooru:
            rawData = mapped.value("posts").toList();
            break;
        default:
            rawData = QVariant(mapped.value("post_ids").toString().split(' ')).toList();
    };
    QVariantMap map = {
        {"id", id},
        {"post_count", postCount},
        {"name", name},
        {"description", description},
        {"raw_post_data", rawData}
    };
    
    return map;
    
}

QVariantMap extractPostData(const QVariant &data, ApiType type) {

    auto mapped = data.toMap();
    
    int id;
    QSet<QString> tags;
    int width;
    int height;
    int fileSize;
    QUrl url;
    QUrl thumbnailUrl;
    QUrl sampleUrl;
    Danbooru::Rating rating;
    
    switch(type) {
        case Danbooru::MoeBooru:
            tags = QSet<QString>::fromList(mapped.value("tags").toString().split(' '));
            id = mapped.value("id").toString().toInt();
            height = mapped.value("height").toString().toInt();
            width = mapped.value("width").toString().toInt();
            url = QUrl::fromUserInput(mapped.value("file_url").toString());
            thumbnailUrl = QUrl::fromUserInput(mapped.value("preview_url").toString());
            sampleUrl = QUrl::fromUserInput(mapped.value("sample_url").toString());
            rating = RATING_MAP.value(mapped.value("rating").toString());
            fileSize = mapped.value("file_size").toInt();
            break;
        case Danbooru::Gelbooru:
            id = 0;
            tags = QSet<QString>();
            width = 0;
            height = 0;
            fileSize = 0;
            url = QUrl();
            thumbnailUrl = QUrl();
            sampleUrl = QUrl();
            rating = Danbooru::Safe;
            break;
        default:
            auto tagKey = mapped.contains("tag_string") ? QL1S("tag_string"): QL1S("tag-string");
            auto widthKey = mapped.contains("image_width") ? QL1S("image_width"): QL1S("image-width");
            auto heightKey = mapped.contains("image_height") ? QL1S("image_height"): QL1S("image-height");
            auto fileSizeKey = mapped.contains("file_size") ? QL1S("file_size"): QL1S("file-size");
            auto fileUrlKey = mapped.contains("file_url") ? QL1S("file_url"): QL1S("file-url");
            auto thumbnailUrlKey = mapped.contains("preview_file_url") ? QL1S("preview_file_url"): QL1S("preview-file-url");
            auto sampleUrlKey = mapped.contains("large_file_url") ? QL1S("large_file_url"): QL1S("large-file-url");
            
            id = mapped.value("id").toString().toInt();
            tags = QSet<QString>::fromList(mapped.value(tagKey).toString().split(' '));
            width = mapped.value(widthKey).toString().toInt();
            height = mapped.value(heightKey).toString().toInt();
            url = mapped.value(fileUrlKey).toUrl();
            thumbnailUrl = mapped.value(thumbnailUrlKey).toUrl();
            fileSize = mapped.value(fileSizeKey).toInt();
            rating = RATING_MAP.value(mapped.value("rating").toString());
            sampleUrl = mapped.value(sampleUrlKey).toUrl();
            break;
    };
    
        
    QVariantMap map = {
        {QL1S("width"), width}, {QL1S("height"), height}, {QL1S("id"), id}, 
        {QL1S("url"), url}, {QL1S("thumbnailUrl"), thumbnailUrl},
        {QL1S("sampleUrl"), sampleUrl}, {QL1S("rating"), rating}, 
        {QL1S("size"), fileSize}, {QL1S("tags"), QVariant::fromValue(tags)}
    };
    
    return(map);
    
}

QList<QVariantMap> parseResult(const QByteArray &data, ApiType type, Danbooru::Request request, bool *result) {
    
    QVariantMap map;
    bool ok;
    QVariant rawMapped = parseDanbooruResult(data, &ok);
    QList<QVariantMap> postList;
    
    if (!ok) {
        *result = false;
        return QList<QVariantMap>();
    }
    
    QVariantList mapped = rawMapped.toList();
    
    // For single queries like pools
    if (mapped.isEmpty()) {
        mapped.append(rawMapped.toMap());
    } 
    
    for (const auto element : mapped) {
        
        QVariantMap mapped;
        
        switch(request) {
            case Pool:
                mapped = extractPoolData(element, type);
                break;
            case Artist:
                qDebug() << "Reimplement";
                break;
            case Tag:
                qDebug() << "Reimplement";
                break;
            default:
                mapped = extractPostData(element, type);
        }
        
        postList.append(mapped);
    }
    
    *result = true; 
    
    return postList;
}

QList<QVariantMap> parseResult(const QXmlStreamAttributes &data, Danbooru::ApiType type, bool *result) {
    return QList<QVariantMap>();
}

} // namespace Danbooru
